/*********************************************************************
** AreaBuilder is a program designed to build maps for MUDS like    **
** Anatolia.                                                        **
** Copyright (C) 2003 Pierre Chatelier                              **
**                                                                  **
** This program is free software; you can redistribute              **
** it and/or modify it under the terms of the GNU                   **
** General Public License as published by the                       **
** Free Software Foundation, version 2.                             **
**                                                                  **
** This program is distributed in the hope that it will be          **
** useful, but WITHOUT ANY WARRANTY; without even the implied       **
** warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. **
** See the GNU General Public License for more details.             **
**                                                                  **
** You should have received a copy of the                           **
** GNU General Public License along with this program; if not,      **
** write to the Free Software Foundation,                           **
** Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA      **
*********************************************************************/
package areabuilder;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.util.*; //iterator
import java.io.*; //for archivation

public class AreaMap extends JPanel implements MouseListener, MouseMotionListener, KeyListener,
                                               Serializable
{
  MainScreen mainScreen = null;
  AnatoliaRoomDialog anatoliaRoomDialog = null;
  AnatoliaMobileDialog anatoliaMobileDialog = null;
  AnatoliaObjectDialog anatoliaObjectDialog = null;
  FindReplaceStringDialog findReplaceStringDialog = null;

  ObjectTreeDialog objectTreeDialog = null;
  ObjectTree objectTree = null;
  LayersDialog layersDialog = null;

  String version = "";
  int areaNumber = 99;
  int minLevel = 1;
  int maxLevel = 99;
  String areaName = "Test area";
  String author = "unknown";
  String resetMessage = "";
  String helpMessage  = "";

  static final float SCALEMIN = .2f;
  static final float SCALEMAX = 3.f;

  static final boolean TOPLEFTCORNER  = false;
  static final boolean REFERENCEPOINT = true;

  PointerState pointerStateAccess = null;
  RoomCollection theRooms = null;
  WireCollection theWires = null;
  MobileCollection theMobiles= null;
  ObjectCollection theObjects= null;
  SelectionOfRooms selectedRooms = null;
  SelectionOfRooms roomClipboard = null;

  int i=0;
  boolean magneticGrid = true;

  Wire currentWire  = null;
  Wire currentWire2 = null;
  Wire wireBeingDrawn = null;
  Room currentRoom = null;
  Room roomDragged = null;
  Room roomCandidateToWire = null;
  Room roomToWire1 = null;
  Room roomToWire2 = null;
  Point corner1 = new Point(0,0);
  Point corner2 = new Point(0,0);
  Point delta = new Point(0,0);

  Dimension size = null;
  float scale = 1.f;

  boolean isDragging = false;
  boolean beginningOfDragging = true;

  boolean isWiring = false;

  boolean isCloning =false;

  int mouseX = 0;
  int mouseY = 0;

  SortedSet selectedLayers = new TreeSet();

  //constructeur
  public AreaMap(MainScreen mainScreen)
  {
    this.mainScreen = mainScreen;
    objectTree = new ObjectTree(this);
    objectTreeDialog = new ObjectTreeDialog(this.mainScreen,"Tree of Objects", false, this.objectTree);
    layersDialog = new LayersDialog(this.mainScreen,"Showing layers...",true,this);
    anatoliaRoomDialog   = new AnatoliaRoomDialog(objectTreeDialog,   "Editing a Room"  , true, this);
    anatoliaMobileDialog = new AnatoliaMobileDialog(objectTreeDialog, "Editing a Mobile", true);
    anatoliaObjectDialog = new AnatoliaObjectDialog(objectTreeDialog, "Editing an Object", true, this);
    findReplaceStringDialog = new FindReplaceStringDialog(this, "Find...", true);

    addKeyListener(this);

    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    size = screenSize;
    setPreferredSize(size);

    theRooms = new RoomCollection(this);
    theWires = new WireCollection();
    theMobiles = new MobileCollection(this);
    theObjects = new ObjectCollection(this);
    selectedRooms = new SelectionOfRooms();

    setBackground(Color.lightGray);
    addMouseListener(this);
    addMouseMotionListener(this);
  }

  public void validateSize()
  {
     this.setMinimumSize(size);
     this.setMaximumSize(size);
     setPreferredSize(size);
     revalidate();
  }

  public void clear()
  {
     LinkingAndSelectableNodeObject.staticClear();
     objectTreeDialog.clear();
     objectTree.clear();
     theRooms.clear();
     theWires.clear();
     theObjects.clear();
     selectedRooms.clear();
     currentWire = null;
     wireBeingDrawn = null;
     currentRoom = null;
     roomDragged = null;
     roomCandidateToWire = null;
     roomToWire1 = null;
     roomToWire2 = null;
     corner1.move(0,0);
     corner2.move(0,0);
     delta.move(0,0);
     selectedLayers.clear();
     Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
     size = screenSize;
     setPreferredSize(size);
     revalidate();
     System.gc();
  }

  public void setPointerStateCursor(PointerState pointerStateAccess)
  {
    this.pointerStateAccess = pointerStateAccess;
  }

  public void setMagneticGrid(boolean state) {magneticGrid = state; repaint();}
  public boolean getMagneticGrid() {return magneticGrid;}

  //dessin
  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);

    if (magneticGrid)
    {
      int step = (int) (2*this.scale*Room.SIZE);
      int gap = (int) (3*scale);
      for(int i=0 ; i<this.getWidth() ; i += step)
      {
        for(int j=0 ; j < this.getHeight() ; j+=step )
        {
           g.drawLine(i,j-gap,i,j+gap);
           g.drawLine(i-gap,j,i+gap,j);
        }
      }
    }

    theRooms.paint(g);
    if(   (pointerStateAccess.getPointerState() == PointerState.WIRESTATE)
       && (currentRoom != null)
      )
    {
        currentRoom.highlightDoor(g,mouseX, mouseY);
    }

    theWires.paint(g);
    if (wireBeingDrawn != null) wireBeingDrawn.paint(g);
    if ( ( corner1.getX() != corner2.getX() ) ||
         ( corner1.getY() != corner2.getY() )
       )
    {
       int x1 = Math.min(corner1.x, corner2.x);
       int y1 = Math.min(corner1.y, corner2.y);
       int width  = Math.abs(corner1.x-corner2.x);
       int height = Math.abs(corner1.y-corner2.y);
       g.drawRect(x1, y1, width, height);
    }

    //avoid blinking while drawing
    if (isWiring)
    {
       Point point = wireBeingDrawn.getLastPoint();
       g.drawLine(point.x, point.y, mouseX, mouseY);
    }
  }

  //change room position if magnetic grid
  public void magneticRooms()
  {
    if (magneticGrid)
    {
      Iterator it = theRooms.getIterator();
      int x;
      int y;
      while(it.hasNext())
      {
        Room room = (Room) it.next();
        x = Math.max(1,
                     Math.round(
                                (
                                 (float) room.getX()
                                 +Room.SIZE/2
                                )
                                /(2*Room.SIZE)
                               )
                    )*(2*Room.SIZE)-Room.SIZE/2;
        y = Math.max(1,
                     Math.round(
                                (
                                 (float) room.getY()
                                 +Room.SIZE/2
                                )
                                /(2*Room.SIZE)
                               )
                    )*(2*Room.SIZE)-Room.SIZE/2;
        room.setLocation(x,y);
      }
    }
  }

  public void fitScreen()
  {
     if (theRooms.getXMaxOnScreen() >= size.width)
     {
       size.width  = theRooms.getXMaxOnScreen()+(int) (Room.SIZE*scale);
     }
     if (theRooms.getYMaxOnScreen() >= size.height)
     {
        size.height = theRooms.getYMaxOnScreen()+(int) (Room.SIZE*scale);
     }

     int THRESHOLD = (int) (50*scale);
     int leftMargin   = (theRooms.getXMinOnScreen() > THRESHOLD            ) ? 0 :
                         Math.abs(theRooms.getXMinOnScreen())+(int) (Room.SIZE*scale);
     int rightMargin  = (theRooms.getXMaxOnScreen() < size.width-THRESHOLD ) ? 0 : 2*THRESHOLD;
     int topMargin    = (theRooms.getYMinOnScreen() > THRESHOLD            ) ? 0 :
                         Math.abs(theRooms.getYMinOnScreen())+(int) (Room.SIZE*scale);
     int bottomMargin = (theRooms.getYMaxOnScreen() < size.height-THRESHOLD) ? 0 : 2*THRESHOLD;

     size.width  += (leftMargin + rightMargin );
     size.height += (topMargin  + bottomMargin);

     Rectangle r = getVisibleRect();
     setPreferredSize(size);
     r.width  = getVisibleRect().width;
     r.height = getVisibleRect().height;
     if (r.width  == 0) r.width  = size.width-leftMargin-rightMargin;
     if (r.height == 0) r.height = size.height-topMargin-bottomMargin;
     r.x += leftMargin;
     r.y += topMargin;
     scrollRectToVisible(r);

     if ((leftMargin !=0) || (topMargin != 0))
     {
      theRooms.translate(leftMargin, topMargin);
      Iterator it = selectedRooms.getListOfInternWiresIterator();
      while(it.hasNext()) ((Wire) it.next()).translate(leftMargin, topMargin);
     }

     this.magneticRooms();

     revalidate();
  }

  public void reScale(float newScale, Point reference, boolean typeOfReference)
  {
    float factor = newScale/scale;
    Rectangle r = getVisibleRect();

    size.width  = (int) (size.width  * factor);
    size.height = (int) (size.height * factor);
    setPreferredSize(size);
    revalidate();

    theRooms.reScale(newScale);
    theWires.reScale(newScale);
    scale = newScale;

    if (typeOfReference == REFERENCEPOINT)
    {
       r.x = (int) (reference.x*(factor-1) + r.x);
       r.y = (int) (reference.y*(factor-1) + r.y);
    }
    else if (typeOfReference == TOPLEFTCORNER)
    {
       r.x = (int) (reference.x*factor)+10;
       r.y = (int) (reference.y*factor)+10;
    }

    scrollRectToVisible(r);
    repaint();
  }

  private void selectRoomsInRectangleOnScreen(Rectangle r)
  {
     Room room;
     selectedRooms.clear();
     Iterator it = theRooms.getIterator();
     while (it.hasNext())
     {
        room = (Room) it.next();
        if (room.isContainedInRectangleOnScreen(r))
           selectedRooms.addToSelection(room);
     }
     mainScreen.menuEditionProperties.setEnabled(!selectedRooms.isEmpty());
     repaint();
  }

  public void mouseEntered(MouseEvent e)
  {
    setCursor(pointerStateAccess.getPointerStateCursor());
  }
  public void mouseExited(MouseEvent e)
  {
      setCursor(pointerStateAccess.getPointerStateDefaultCursor());
      if (isWiring)
      {
        isWiring = false;
        wireBeingDrawn = null;
        repaint();
      }
  }

  public void mousePressed(MouseEvent e)
  {
     this.requestFocus();
     if (pointerStateAccess.getPointerState() == PointerState.POINTERSTATE)
     {
        roomDragged = theRooms.inWhichRoom(e.getX(), e.getY(), scale);

        //delta is the shift between the (x,y) of the room and the real point
        //inside that one clicked on
        if (roomDragged != null)
        {
          isDragging = true;
          beginningOfDragging = true;
          delta.move(e.getX()-roomDragged.getXOnScreen(),e.getY()-roomDragged.getYOnScreen());
        }
        else if (roomDragged == null)
        {
           isDragging = true;
           corner1.setLocation(e.getPoint());
           corner2.setLocation(e.getPoint());
           corner2.setLocation(e.getPoint());
        }
     }
     else if (pointerStateAccess.getPointerState() == PointerState.ZOOMSTATE)
     {
        if (!isDragging)
        {
          corner1.setLocation(e.getPoint());
          corner2.setLocation(e.getPoint());
          isDragging = true;
        }
        else
          corner2.setLocation(e.getPoint());
     }
  }
  public void mouseReleased(MouseEvent e)
  {
    if (pointerStateAccess.getPointerState() == PointerState.POINTERSTATE)
    {
       if (roomDragged == null) //selecting rooms by drawing a rectangle
       {
          if (isDragging)
          {
             if (!(corner2.equals(corner1)))
             {
                int x1 = Math.min(corner1.x, corner2.x);
                int y1 = Math.min(corner1.y, corner2.y);
                int width  = Math.abs(corner1.x-corner2.x);
                int height = Math.abs(corner1.y-corner2.y);
                Rectangle r = new Rectangle(x1,y1,width,height);
                selectRoomsInRectangleOnScreen(r);
             }
             isDragging = false;
          }
          else if (!isDragging)
          {
          }
          corner1.move(0,0);
          corner2.move(0,0);
          }
       // if we tried to move a room...
       else if ((isDragging) && (roomDragged != null) && (roomDragged.hasMoved()))
       {
         //move it!
          if (selectedRooms.isEmpty())
          {
             if (isCloning)
               mainScreen.performUndoableAction(new UndoableActionAreaMapRoomMove(this, roomDragged, UndoableAction.AUTOMATIC_ACTION));
             else
               mainScreen.performUndoableAction(new UndoableActionAreaMapRoomMove(this, roomDragged));
          }
          else
          {
             if (isCloning)
               mainScreen.performUndoableAction(new UndoableActionAreaMapGroupRoomMove(this, selectedRooms, UndoableAction.AUTOMATIC_ACTION));
             else
               mainScreen.performUndoableAction(new UndoableActionAreaMapGroupRoomMove(this, selectedRooms));
          }
          theRooms.resetMinMax();
          fitScreen();
          repaint();
          roomDragged = null;
          isDragging = false;
          isCloning = false;
       }
    }
    else if (pointerStateAccess.getPointerState() == PointerState.ZOOMSTATE)
    {
      isDragging = false;
      if (!(corner2.equals(corner1)))
      {
         int x1 = Math.min(corner1.x, corner2.x);
         int y1 = Math.min(corner1.y, corner2.y);
         int width  = Math.abs(corner1.x-corner2.x);
         int height = Math.abs(corner1.y-corner2.y);
         float newScale=1.f;
         if (width < height)
            newScale = (getVisibleRect().width)/(width/scale);
         else
            newScale = (getVisibleRect().height)/(height/scale);
         newScale = Math.min(SCALEMAX, newScale);
         reScale(newScale,new Point(x1,y1), TOPLEFTCORNER);
      }
      corner1.move(0,0);
      corner2.move(0,0);
    }
  }

  public void mouseClicked(MouseEvent e)
  {
     if (pointerStateAccess.getPointerState() == PointerState.POINTERSTATE)
     {
        //if  we clicked on a new room
        if (roomDragged != null)
        {
           if (((e.getModifiers() & e.BUTTON3_MASK) != 0) ||  (e.getClickCount() == 2))
           {
              selectedRooms.removeFromSelection(roomDragged);
              anatoliaRoomDialog.showFor(roomDragged);
           }
           else
           {
              if (e.isControlDown())
                 selectedRooms.toggleInSelection(roomDragged);
              else
                 selectedRooms.resetSelection(roomDragged);

              TreePath[] tab = new TreePath[selectedRooms.getSize()];
              Iterator it = selectedRooms.getIterator();
              for(int i=0 ; i<tab.length ; ++i)
              {
                tab[i] = new TreePath (((Room) it.next()).anatoliaRoom.getPath());
              }
              objectTree.treeOfObjects.setSelectionPaths(tab);
              if (tab.length>0) objectTree.treeOfObjects.scrollPathToVisible(tab[0]);

             mainScreen.menuEditionProperties.setEnabled(!selectedRooms.isEmpty());
           }
        }
        else//clear the whole selection
        {
           selectedRooms.clearSelection();
           mainScreen.menuEditionProperties.setEnabled(!selectedRooms.isEmpty());
        }

        //did we click on a wire?
        theWires.unselectAll();
        currentWire = theWires.getWireGoingTrough(e.getPoint());
        if (currentWire !=  null)
        {
           currentWire.setAllWireSelected(e.getClickCount()>=2);
           Graphics g = getGraphics();
           currentWire.paint(g);
           g.dispose();
        }
        else theWires.unselectAll();
        repaint();
     }
     else if (pointerStateAccess.getPointerState() == PointerState.WIRESTATE)
     {
        if (isWiring)
        {
           if (currentRoom != null)// && (currentRoom != roomToWire1))
           {
                roomToWire2 = currentRoom;
                if (wireBeingDrawn.close(e.getX(), e.getY(), roomToWire2))
                {
                   mainScreen.performUndoableAction(new UndoableActionAreaMapWireAdd(this, wireBeingDrawn));
                   wireBeingDrawn = null;
                   isWiring = false;
                }
           }
           else
           {
              wireBeingDrawn.add(e.getPoint());
           }
        }
        else
        {
           if (currentRoom != null)
           {
              wireBeingDrawn = new Wire();
              roomToWire1 = currentRoom;
              if (wireBeingDrawn.open(e.getX(), e.getY(), roomToWire1, scale))
              {
                 isWiring = true;
              }
           }
           else if (wireBeingDrawn != null) wireBeingDrawn.add(e.getPoint());
        }
     }
     else if (pointerStateAccess.getPointerState() == PointerState.ZOOMSTATE)
     {
        if (((e.getModifiers() & e.BUTTON1_MASK) != 0) || ((e.getModifiers() & e.BUTTON3_MASK) != 0))
        {
          float newScale = 1.f;
          if      ((e.getModifiers() & e.BUTTON1_MASK) != 0) newScale = Math.min(scale*1.5f, SCALEMAX);
          else if ((e.getModifiers() & e.BUTTON3_MASK) != 0) newScale = Math.max(scale/1.5f, SCALEMIN);

          reScale(newScale, e.getPoint(), REFERENCEPOINT);
        }
     }
     else if (pointerStateAccess.getPointerState() == PointerState.ROOMSTATE)
     {
       int x = e.getX();
       int y = e.getY();
       if (magneticGrid)
       {
         x = Math.max(1,Math.round( ((float) x)/(2*Room.SIZE) ))*(2*Room.SIZE)-Room.SIZE/2;
         y = Math.max(1,Math.round( ((float) y)/(2*Room.SIZE) ))*(2*Room.SIZE)-Room.SIZE/2;
       }
       if ((e.getModifiers() & e.BUTTON1_MASK) != 0)
       {
          if (selectedLayers.isEmpty() && !theRooms.isEmpty())
          {
            JOptionPane.showMessageDialog(this.mainScreen,
                                         "No layers are visible,"+"\n"+"you cannot create a room",
                                         "Error",JOptionPane.WARNING_MESSAGE);
          }
          else
          {
            int layer = 0;
            if (selectedLayers.isEmpty())
            {
              layer = 0;
              selectedLayers.add(new Integer(layer));
            }
            else
            {
              layer = ((Integer) selectedLayers.first()).intValue();
            }
            Room room = new Room(this, x, y, scale, theRooms.anatoliaCollection);
            room.layer = layer;
            mainScreen.performUndoableAction(new UndoableActionAreaMapRoomAdd(this, room));
            Graphics g = getGraphics();
            room.paint(g);
            g.dispose();
            fitScreen();
          }
       }
     }
     repaint();
  }

  public void mouseDragged(MouseEvent e)
  {
    mouseX = e.getX();
    mouseY = e.getY();

    if (pointerStateAccess.getPointerState() == PointerState.POINTERSTATE)
    {
       if (roomDragged == null) //selecting rooms by drawing a rectangle
       {
           corner2.setLocation(e.getPoint());
           if (!(corner2.equals(corner1)))
           {
               repaint();
               Graphics g = getGraphics();
               int x1 = Math.min(corner1.x, corner2.x);
               int y1 = Math.min(corner1.y, corner2.y);
               int width  = Math.abs(corner1.x-corner2.x);
               int height = Math.abs(corner1.y-corner2.y);
               g.drawRect(x1, y1, width, height);
               g.dispose();
           }
       }
       else if (roomDragged != null) //dragging a room
       {
         if (!selectedRooms.contains(roomDragged))
         {
            selectedRooms.clearSelection();
         }
         if (beginningOfDragging && e.isControlDown())
         {
            beginningOfDragging = false;
            if (selectedRooms.isEmpty())
            {
               mainScreen.performUndoableAction(
                   new UndoableActionAreaMapRoomClone(this, roomDragged)
               );
               roomDragged.translating((e.getX()-roomDragged.getCandidateXOnScreen())-delta.x,
                                       (e.getY()-roomDragged.getCandidateYOnScreen())-delta.y);
               isCloning = true;
            }
            else
            {
               mainScreen.performUndoableAction(
                  new UndoableActionAreaMapGroupRoomClone(this, selectedRooms)
               );
               selectedRooms.translating((e.getX()-roomDragged.getCandidateXOnScreen())-delta.x,
                                         (e.getY()-roomDragged.getCandidateYOnScreen())-delta.y);
               isCloning = true;
            }
         }

         //delta is the shift between the (x,y) of the room and
         //the real point inside one clicked on
         if (selectedRooms.isEmpty())
         {
            roomDragged.translating((e.getX()-roomDragged.getCandidateXOnScreen())-delta.x,
                                    (e.getY()-roomDragged.getCandidateYOnScreen())-delta.y);
         }
         else
         {
            selectedRooms.translating((e.getX()-roomDragged.getCandidateXOnScreen())-delta.x,
                                      (e.getY()-roomDragged.getCandidateYOnScreen())-delta.y);
         }
         repaint();
       }
    }
    else if (pointerStateAccess.getPointerState() == PointerState.ZOOMSTATE)
    {
      corner2.setLocation(e.getPoint());
      if (!(corner2.equals(corner1)))
      {
        repaint();
        Graphics g = getGraphics();
        int x1 = Math.min(corner1.x, corner2.x);
        int y1 = Math.min(corner1.y, corner2.y);
        int width  = Math.abs(corner1.x-corner2.x);
        int height = Math.abs(corner1.y-corner2.y);
        g.drawRect(x1, y1, width, height);
        g.dispose();
      }
    }
  }

  public void mouseMoved(MouseEvent e)
  {
      mouseX = e.getX();
      mouseY = e.getY();

      if (pointerStateAccess.getPointerState() == PointerState.WIRESTATE)
      {
        currentRoom = theRooms.inWhichRoom(e.getX(), e.getY(), scale);

        if ((currentRoom != roomCandidateToWire) && (roomCandidateToWire != null))
        {
            Graphics g = getGraphics();
            roomCandidateToWire.paint(g);
            g.dispose();
        }

        if (currentRoom != null)
        {
            roomCandidateToWire = currentRoom;
            Graphics g = getGraphics();
            roomCandidateToWire.highlightDoor(g, e.getX(), e.getY());
            g.dispose();
        }

        if (isWiring)
        {
            repaint();
        }
      }
   }

   public void keyReleased(KeyEvent e) {}

   public void keyPressed(KeyEvent e)
   {
      if (e.getKeyChar() == e.VK_SPACE) //toggle wire/pointer mode
      {
         if (pointerStateAccess.pointerState == PointerState.POINTERSTATE)
         {
            mainScreen.wireButton.setSelected(true);
            pointerStateAccess.setPointerState(PointerState.WIRESTATE);
            setCursor(pointerStateAccess.getPointerStateCursor());
         }
         else if (pointerStateAccess.pointerState == PointerState.WIRESTATE)
         {
            mainScreen.pointerButton.setSelected(true);
            pointerStateAccess.setPointerState(PointerState.POINTERSTATE);
            setCursor(pointerStateAccess.getPointerStateCursor());
         }
      } //deletion
      else if (e.getKeyCode() == e.VK_DELETE)
      {
         if (pointerStateAccess.getPointerState() == PointerState.POINTERSTATE)
         {
            //DELETE deletes a room
            if (!selectedRooms.isEmpty())
            {
               mainScreen.performUndoableAction(
                  new UndoableActionAreaMapGroupRoomRemove(this, selectedRooms)
               );
            }
            else if (currentWire != null)
            {
               mainScreen.performUndoableAction(
                   new UndoableActionAreaMapWireSegmentDelete(this, currentWire)
               );
            }
            repaint();
         }
      }
      else if (e.getKeyCode() == e.VK_ESCAPE)
      {
         if (pointerStateAccess.getPointerState() == PointerState.WIRESTATE)
         {
            //stops wiring
            isWiring = false;
            wireBeingDrawn = null;
            repaint();
         }
      }
   }
   public void keyTyped(KeyEvent e)
   {
      if (pointerStateAccess.getPointerState() == PointerState.WIRESTATE)
      {
         if (isWiring && (wireBeingDrawn != null))
         {
            if (e.getKeyChar() == e.VK_BACK_SPACE)
            {
               if (wireBeingDrawn.size() == 0)
               {
                  isWiring = false;
                  wireBeingDrawn = null;
               }
               else
               {
                 wireBeingDrawn.removeLastPoint();
               }
               repaint();
            }
         }
      }
   }

   public void buildArea(FileWriter output) throws IOException
   {
     output.write("#AREA\n");
     String name = new String(mainScreen.areaFileName);
     name = name.substring(name.lastIndexOf("/")+1);
     name = name.substring(name.lastIndexOf("\\")+1);
     int pos = name.lastIndexOf(".");
     if (pos>=0) name = name.substring(0,pos);
     output.write(name+".are"+"~"+"\n");
     output.write(this.areaName+"~"+"\n");
     output.write("{"+this.minLevel+" "+this.maxLevel+"} "+this.author+" "+this.areaName+"~"+"\n");
     int maxVnum = Math.max(theMobiles.getMaxVnum(),theRooms.anatoliaCollection.getMaxVnum());
     maxVnum = (maxVnum/100)*100+99;
     output.write(this.areaNumber*1000+" "+maxVnum+"\n");
     output.write("\n");
     output.write("\n");

     if (!resetMessage.trim().equals(""))
     {
       output.write("#RESETMESSAGE\n");
       output.write(resetMessage+"~"+"\n");
       output.write("\n");
     }
     if (!helpMessage.trim().equals(""))
     {
       output.write("#HELPS\n");
       output.write(helpMessage+"\n"+"0 $~"+"\n");
       output.write("\n");
     }

     output.write("#ROOMS\n");
     theRooms.buildRoomDescriptions(output);
     output.write("#0\n");
     output.write("\n");

     output.write("#MOBILES\n");
     theMobiles.buildMobileDescriptions(output);
     output.write("#0\n");
     output.write("\n");

     output.write("#OBJECTS\n");
     theObjects.buildObjectDescriptions(output);
     output.write("#0\n");
     output.write("\n");

     output.write("#RESETS\n");
     theRooms.buildRoomResets(output);
     theObjects.buildObjectResets(output);
     theMobiles.buildMobileResets(output);
     output.write("S\n");
     output.write("\n");

     output.write("#SHOPS\n");
     theMobiles.buildMobileShops(output);
     output.write("0\n");
     output.write("\n");

     output.write("#SPECIALS\n");
     theMobiles.buildMobileSpecials(output);
     output.write("S\n");
     output.write("\n");

     output.write("#OMPROGS\n");
     theMobiles.buildMobileProgs(output);
     theObjects.buildObjectProgs(output);
     output.write("S\n");
     output.write("\n");

     output.write("#OLIMITS\n");
     theObjects.buildObjectLimits(output);
     output.write("S\n");
     output.write("\n");

     output.write("#$\n");
    }

  public void writeObject(ObjectOutputStream output) throws IOException
   {
     output.writeInt(areaNumber);
     output.writeInt(minLevel);
     output.writeInt(maxLevel);
     output.writeUTF(areaName);
     output.writeUTF(author);
     output.writeUTF(resetMessage);
     output.writeUTF(helpMessage);

     output.writeInt(size.width);
     output.writeInt(size.height);
     theRooms.writeObject(output);
     theWires.writeObject(output);
     theMobiles.writeObject(output);
     theObjects.writeObject(output);
     objectTree.writeObject(output);
   }

   public void readObject(ObjectInputStream input) throws IOException, ClassNotFoundException
   {
      areaNumber   = input.readInt();
      minLevel     = input.readInt();
      maxLevel     = input.readInt();
      areaName     = input.readUTF();
      author       = input.readUTF();
      resetMessage = input.readUTF();
      helpMessage  = input.readUTF();

      size.setSize(input.readInt(), input.readInt());
      theRooms = new RoomCollection(this);
      theRooms.readObject(input);
      theWires = new WireCollection();
      theWires.readObject(input);
      theMobiles = new MobileCollection(this);
      theMobiles.readObject(input);
      theObjects = new ObjectCollection(this);
      theObjects.readObject(input);
      objectTree.readObject(input); //establish some links between instances

      theRooms.establishLinks();
      theWires.establishLinks();
      theMobiles.establishLinks();
      theObjects.establishLinks();
      LinkingAndSelectableNodeObject.defragIDs();

      objectTreeDialog.pack();
      objectTreeDialog.repaint();
      repaint();
   }

   public void act_copy()
   {
      roomClipboard = (SelectionOfRooms) selectedRooms.clone();
   }

   public void act_cut()
   {
      act_copy();
      mainScreen.performUndoableAction(
           new UndoableActionAreaMapGroupRoomRemove(this, selectedRooms)
      );
      repaint();
   }
   public void act_paste()
   {
      if (roomClipboard == null) return;
      mainScreen.performUndoableAction(
           new UndoableActionAreaMapGroupRoomClone(this, roomClipboard));
      selectedRooms.translating((int) (Room.SIZE*scale),(int) (Room.SIZE*scale));
      mainScreen.performUndoableAction(
           new UndoableActionAreaMapGroupRoomMove(this, selectedRooms, UndoableAction.AUTOMATIC_ACTION));
      act_copy();
      theRooms.resetMinMax();
      fitScreen();
      repaint();
   }

   public void setAreaNumber(int newAreaNumber)
   {
      if ((newAreaNumber<1) || (newAreaNumber>99)) return;
      theRooms.changeAreaNumber(areaNumber, newAreaNumber);
      theMobiles.changeAreaNumber(areaNumber, newAreaNumber);
      areaNumber = newAreaNumber;
   }
}